(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_graph.ml,v 1.10 2013-12-02 15:28:38 fclement Exp $ *)

(* Matrix graphs of function. *)

module Nat = Unatural;;

module Idx = Uindex_index;;
module Set = Umatrix_nonempty_set;;

let mg_invalid_arg = Upervasives.invalid_arg "Ugraph";;

type index = Set.index
and uindex = Set.uindex
;;

type domain = Set.t;;

type sub_domain = Set.nonempty_subset;;

type 'a images =
  | Constant of 'a
  | Images of 'a array
;;

type 'a t =
  {
   domain : domain;
   images : 'a images;
  }
;;

let same mg1 mg2 = mg1 == mg2;;

(* Construction functions. *)

let make m =
  let sz = Umatrix.size m in
  let dom = Set.make sz in
  let ims = Umatrix.to_array m in
  { domain = dom; images = Images ims; }
;;

let constant dom y = { domain = dom; images = Constant y; };;

let from_array dom ims =
  if Set.cardinal dom <> Uarray.length ims then
    mg_invalid_arg "from_array" "arguments have different sizes" else
  { domain = dom; images = Images ims; }
;;

let from_fun dom f =
  let card = Set.cardinal dom in
  let (idx0, uidx0) = Set.choice dom in
  let y0 = f idx0 uidx0 in
  let ims = Uarray.make card y0 in
  let constant = ref true in
  let ff idx uidx =
    let i = Idx.to_int idx in
    let y = f idx uidx in
    if y = y0 then ims.(i) <- y0 else
    begin constant := false; ims.(i) <- y end in
  Set.sub_iter ff dom Idx._1_ (Nat.pred card);
  let images = if !constant then Constant y0 else Images ims in
  { domain = dom; images = images; }
;;

(* Access functions. *)

let get_constant mg =
  match mg.images with
  | Constant y -> y
  | Images _ -> mg_invalid_arg "get_constant" "non constant graph"
;;

let image mg =
  match mg.images with
  | Constant y -> (fun _idx -> y)
  | Images ims -> (fun idx -> ims.(Idx.to_int idx))
;;

let restriction
    ({ domain = dom; images = images; }  as mg)
    { Set.support = sup; indices = indices; nonempty_set = sub_dom; } =
  if sup != dom then
    mg_invalid_arg "restriction" "sub-domain is not a subset of domain" else
  match images with
  | Constant y -> constant sub_dom y
  | Images _ -> from_array sub_dom (Array.map (image mg) indices)
;;

(* Iterators. *)

let map_images f mg =
  match mg.images with
  | Constant y -> { mg with images = Constant (f y); }
  | Images ims -> { mg with images = Images (Array.map f ims); }
;;
