(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: udebug.ml,v 1.29 2014/01/10 15:28:04 fclement Exp $ *)

(* Debug/report/error message facilities. *)

type purpose = | User | Library
and software_name = string
and tag = string
and 'a dynamic_tag = 'a -> tag
and prefix = string
;;

type short_key_prefix =
  | Automatic
  | Short of prefix

and key_prefixes =
  {
    long : prefix;
    short : short_key_prefix;
  }
;;

type 'a printing_facilities =
  {
   report : Upervasives.message -> bool;
   report_action : (unit -> Upervasives.message) -> bool;
   debug : Upervasives.function_name -> Upervasives.message -> bool;
   debug_action :
     Upervasives.function_name -> (unit -> Upervasives.message) -> bool;
   fatal_error :
     Upervasives.function_name -> Upervasives.message -> int -> 'a;
  }

and ('a, 'b) dynamic_printing_facilities =
  {
   dyn_report : 'a -> Upervasives.message -> bool;
   dyn_report_action : 'a -> (unit -> Upervasives.message) -> bool;
   dyn_debug : 'a -> Upervasives.function_name -> Upervasives.message -> bool;
   dyn_debug_action :
     'a -> Upervasives.function_name -> (unit -> Upervasives.message) -> bool;
   dyn_fatal_error :
     'a -> Upervasives.function_name -> Upervasives.message -> int -> 'b;
  }
;;

(* Key prefixes. *)

let make_key_prefixes sname =
  { long = Ustring.long_key_of_string sname; short = Automatic; }
;;

(* Printing. *)

let print_announce_opt ob =
  Upervasives.option (Printf.bprintf ob "%s ") ()
;;

let print_ann_opt ob =
  Upervasives.option
    (fun (announce_opt, fun_name) ->
      Printf.bprintf ob "%a%s: " print_announce_opt announce_opt fun_name)
    ()
;;

let print_ann_opt_message ob (ann_opt, msg) =
  Printf.bprintf ob "%a%s" print_ann_opt ann_opt msg
;;

let print_tag_ann_opt_message ob (f, x, ann_opt, msg) =
  let tag = f x in
  let prefix = if tag <> "" then Printf.sprintf "%s " tag else "" in
  Printf.bprintf ob "%s%a" prefix print_ann_opt_message (ann_opt, msg)
;;

(* Messages with dynamic tags. *)

let make_printing_facilities_with_dynamic_tags
    purpose sname { long = long_kp; short = short_kp; } make_tag =
  
  (* Report and debug flags. *)
  
  let (report_flag, debug_flag) = Uaref.make_embedded_flags () in
  
  let add_flag kind k comment set_flag add =
    let adv_cmd_opt =
      Ucommand_option.of_regular
        { Ucommand_option.key =
            Printf.sprintf "-%s-%s" long_kp kind;
          spec = Arg.Unit set_flag;
          doc =
            Printf.sprintf
              "\n\tTurn %s %s mode on. Default is off.%s"
              sname kind comment; } in
    let adv_cmd_opt =
      match short_kp with
      | Automatic ->
          Ucommand_option.with_auto_short_key adv_cmd_opt
      | Short skp ->
          Ucommand_option.with_short_key
            (Printf.sprintf "-%s%c" skp k) adv_cmd_opt in
    add adv_cmd_opt; in
  
  let add_report_flag () =
    let report_add =
      match purpose with
      | User -> Ucommand_option.report_user_add
      | Library -> Ucommand_option.report_library_add in
    add_flag "report" 'k' ""
      report_flag.Uaref.set_flag report_add in
  add_report_flag ();

  let add_debug_flag () =
    let debug_add =
      match purpose with
      | User -> Ucommand_option.debug_user_add
      | Library -> Ucommand_option.debug_library_add in
    add_flag "debug" 'd'
      (Printf.sprintf "\n\tImplies %s report mode." sname)
      debug_flag.Uaref.set_flag debug_add in
  add_debug_flag ();

  (* Printing functions. *)
  
  let ob = Buffer.create 1024 in
  
  let display_message_on_stderr make_tag x ann_opt msg =
    Printf.bprintf ob "%a\n"
      print_tag_ann_opt_message (make_tag, x, ann_opt, msg);
    Ubuffer.flush_buffer_on_stderr ob in
  
  let report_action x action =
    if report_flag.Uaref.get_flag () then begin
      let msg = action () in
      display_message_on_stderr make_tag x None msg
    end;
    true in
  
  let report x msg = report_action x (fun () -> msg) in
  
  let debug_action x fun_name action =
    if debug_flag.Uaref.get_flag () then begin
      let msg = action () in
      display_message_on_stderr make_tag x (Some (None, fun_name)) msg
    end;
    true in
  
  let debug x fun_name msg = debug_action x fun_name (fun () -> msg) in
  
  let fatal_error x fun_name msg error_code =
    let announce = "Fatal error in" in
    display_message_on_stderr make_tag x (Some (Some announce, fun_name)) msg;
    exit error_code in
  
  {
   dyn_report = report;
   dyn_report_action = report_action;
   dyn_debug = debug;
   dyn_debug_action = debug_action;
   dyn_fatal_error = fatal_error;
  }
;;

let gen_make_printing_facilities purpose sname kp make_tag =
  let
      {
       dyn_report = report;
       dyn_report_action = report_action;
       dyn_debug = debug;
       dyn_debug_action = debug_action;
       dyn_fatal_error = fatal_error;
      } =
    make_printing_facilities_with_dynamic_tags purpose sname kp make_tag in
  {
   report = report ();
   report_action = report_action ();
   debug = debug ();
   debug_action = debug_action ();
   fatal_error = fatal_error ();
  }
;;

let make_printing_facilities purpose sname kp =
  let make_tag () = "" in
  gen_make_printing_facilities purpose sname kp make_tag
;;

let make_printing_facilities_with_static_tag purpose sname kp =
  let make_tag () = Printf.sprintf "[%s]" sname in
  gen_make_printing_facilities purpose sname kp make_tag
;;
