(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mapping.ml,v 1.1 2013/06/13 19:01:17 fclement Exp $ *)

let mapping_invalid_arg fun_name msg =
  Pervasives.invalid_arg (Printf.sprintf "Mapping.%s: %s" fun_name msg)
;;

open Patchwork_types;;
open Mapping_types;;

(*
let gen_digest = Utype_digest.gen_binary_digest Mapping_digest.md5;;
let extensional_mapping_digest = gen_digest "extensional_mapping"
and intensional_mapping_digest = gen_digest "intensional_mapping"
and intensional_structural_mapping_digest =
  gen_digest "intensional_structural_mapping"
and extensional_structural_mapping_digest =
  gen_digest "extensional_structural_mapping"
;;
*)

(* Converters. *)

let domain_of_extensional_mapping em = em.Ugraph.domain;;

let sub_extensional_mapping em ss =
  if em.Ugraph.domain != ss.Uset.support then
    mapping_invalid_arg "sub_em"
      "subset is not included in extensional mapping domain" else
  match em.Ugraph.images with
  | Ugraph.Constant y -> Ugraph.constant ss.Uset.set y
  | Ugraph.Images ims ->
      let image index = ims.(Uindex.to_int index) in
      let images = Array.map image ss.Uset.indices in
      Ugraph.from_array ss.Uset.set images
;;

(*
let in_domain dom x =
  match dom with
  | Extensional_domain s -> Uset.mem x s
  | Intensional_domain chi -> chi x
;;
*)
let in_domain dom x = Uset.mem dom x;;

let domain_of_intensional_structural_mapping ism =
  let rec loop accu = function
  | Piece { domain = dom; _; } -> Uset.union accu dom
  | Stitch { patch = ismp; weave = ismw; } -> loop (loop accu ismp) ismw in
  (loop (Uset.empty ()) ism)
;;

(* For each future evaluation of the resulting application, the argument is
 processed through the patchwork to find the Piece that contains it.
 This might be an efficiency issue... *)
let application_of_intensional_structural_mapping ism =
  let rec loop = function
    | Piece { application = f; _; } -> f
    | Stitch { patch = ismp; weave = ismw; } ->
        let dom = domain_of_intensional_structural_mapping ismp in
        (fun x -> if in_domain dom x then loop ismp x else loop ismw x) in
  loop ism
;;

let intensional_mapping_of_intensional_structural_mapping ism =
  let dom = domain_of_intensional_structural_mapping ism in
  let f = application_of_intensional_structural_mapping ism in
  { domain = dom; application = f; }
;;

let intensional_mapping_of_extensional_mapping em =
  let dom = domain_of_extensional_mapping em in
  let f =
    match em.Ugraph.images with
    | Ugraph.Constant y -> (fun _ -> y)
    | Ugraph.Images ims -> (fun uindex -> ims.(Uindex.to_int uindex)) in
  { domain = dom; application = f; }
;;

let intensional_structural_mapping_of_extensional_mapping em =
  Piece (intensional_mapping_of_extensional_mapping em)
;;

let intensional_structural_mapping_of_intensional_mapping im = Piece im;;

let extensional_mapping_of_intensional_mapping
    { domain = dom; application = f; } =
  Ugraph.from_fun dom (fun _ uindex -> f uindex)
;;

let extensional_mapping_of_intensional_structural_mapping ism =
  extensional_mapping_of_intensional_mapping
    (intensional_mapping_of_intensional_structural_mapping ism)
;;

let domain_of_extensional_structural_mapping esm =
  let f accu { Ugraph.domain = dom; _; } = Uset.union accu dom in
  Ulist1.fold_left f (Uset.empty ()) esm
;;

let esm_choice esm =
  let f_item accu em =
    match (accu, em) with
    | (Some _, _)
    | (None, { Ugraph.domain = Uset.Empty; _; }) -> accu
    | (None, { Ugraph.domain = Uset.Nonempty _;
               images = Ugraph.Constant y; }) -> Some y
    | (None, { Ugraph.domain = Uset.Nonempty _;
               images = Ugraph.Images ims; }) -> Some ims.(0) in
  match Ulist1.fold_left f_item None esm with
  | None -> raise Not_found
  | Some choice -> choice
;;

let images_of_extensional_structural_mapping esm = function
  | Uset.Empty -> [||]
  | Uset.Nonempty ne as dom ->
      let y0 = esm_choice esm in
      let ims = Uarray.make (Uset.cardinal dom) y0 in
      let f_item = function
        | { Ugraph.domain = item_dom; images = Ugraph.Constant item_y; } ->
              let f item_uindex =
                let index = Uuniverse.index ne.Uset.uindices item_uindex in
                (*let index = Uset.full_index_of_uindex item_uindex in*)
                ims.(Uindex.to_int index) <- item_y in
              Uset.iter f item_dom
        | { Ugraph.domain = item_dom; images = Ugraph.Images item_ims; } ->
              let f item_index item_uindex =
                let index = Uuniverse.index ne.Uset.uindices item_uindex in
                ims.(Uindex.to_int index) <-
                  item_ims.(Uindex.to_int item_index) in
              Uset.iter_i f item_dom in
      Ulist1.iter f_item esm;
      ims
;;

let extensional_mapping_of_extensional_structural_mapping esm =
  let dom = domain_of_extensional_structural_mapping esm in
  let ims = images_of_extensional_structural_mapping esm dom in
  Ugraph.from_array dom ims
;;

let extensional_structural_mapping_of_extensional_mapping em = (em, []);;

(* Iterators. *)

let extensional_mapping_map f_image em = Ugraph.map_images f_image em;;

let intensional_structural_mapping_fold_left = Patchwork.fold_left;;

let intensional_structural_mapping_map = Patchwork.map;;

let extensional_structural_mapping_fold_left = Ulist1_zipper.fold_left;;

let extensional_structural_mapping_map = Ulist1.map;;
